from collections.abc import Callable
from functools import partial
from threading import Event, Thread
from typing import Any, Self

from confluent_kafka import KafkaException, Producer

from ampel.abstract.AbsContextManager import AbsContextManager
from ampel.base.decorator import abstractmethod

from .KafkaAuthentication import KafkaAuthentication


class KafkaProducerBase[T](AbsContextManager, abstract=True):
    bootstrap: str
    topic: str
    auth: None | KafkaAuthentication = None

    kafka_producer_properties: dict[str, Any] = {}
    delivery_timeout: float = 10.0

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._producer = Producer(
            **{
                "bootstrap.servers": self.bootstrap,
            }
            | (self.auth.librdkafka_config() if self.auth else {})
            | self.kafka_producer_properties
        )
        self._stop_thread = Event()
        self._thread: None | Thread = None

    def _poll(self):
        # Poll producer from a thread to trigger delivery callbacks
        while not self._stop_thread.is_set():
            self._producer.poll(1)

    @abstractmethod
    def serialize(self, message: T) -> bytes: ...  # type: ignore[empty-body]

    def _on_delivery(
        self,
        hook: None | Callable[[], None],
        err,
        msg,  # noqa: ARG002
    ):
        if err is not None:
            raise KafkaException(err)
        if hook is not None:
            hook()

    def produce(self, message: T, delivery_callback: None | Callable[[], None]) -> None:
        self._producer.produce(
            self.topic,
            self.serialize(message),
            on_delivery=partial(self._on_delivery, delivery_callback),
        )

    def __enter__(self) -> "Self":
        assert self._thread is None, f"{self.__class__.__qualname__} is not reentrant"
        # start the delivery callback thread
        self._stop_thread.clear()
        self._thread = Thread(target=self._poll)
        self._thread.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        # stop the delivery callback thread
        assert self._thread is not None
        self._stop_thread.set()
        self._thread.join()
        self._thread = None
        # ensure enqueued messages are delivered
        if (in_queue := self._producer.flush(self.delivery_timeout)) > 0:
            raise TimeoutError(
                f"{in_queue} messages still in queue after {self.delivery_timeout} s"
            )
        # trigger callbacks for any remaining messages
        self._producer.poll(0)



#!/usr/bin/env python
# File:                Ampel-ZTF/ampel/ztf/t3/complement/ZTFCutoutImages.py
# Author:              Jakob van Santen <jakob.van.santen@desy.de>
# Date:                18.09.2020
# Last Modified Date:  18.09.2020
# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>

from base64 import b64decode
from collections.abc import Iterable
from typing import Literal

from requests_toolbelt.sessions import (  # type: ignore[import-untyped]
    BaseUrlSession,
)

from ampel.abstract.AbsBufferComplement import AbsBufferComplement
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.struct.T3Store import T3Store
from ampel.ztf.base.CatalogMatchUnit import retry_transient_errors


class LSSTCutoutImages(AbsBufferComplement):
    """
    Add cutout images from LSST archive database
    """

    #: Which detection to retrieve cutouts for
    eligible: Literal["first", "last", "brightest", "all"] = "last"

    archive_url: str = "https://ampel-dev.ia.zeuthen.desy.de/api/lsst/archive/v1/"
    insecure: bool = False

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self.session = BaseUrlSession(base_url=self.archive_url)

    @retry_transient_errors()
    def get_cutout(self, diaSourceId: int) -> None | dict[str, bytes]:
        response = self.session.get(
            f"alert/{diaSourceId}/cutouts", verify=not self.insecure
        )
        if response.status_code == 404:
            return None

        response.raise_for_status()
        json = response.json()
        return {
            k: b64decode(json[k])
            for k in ["cutoutScience", "cutoutTemplate", "cutoutDifference"]
        }

    def complement(self, records: Iterable[AmpelBuffer], t3s: T3Store) -> None:  # noqa: ARG002
        for record in records:
            if (photopoints := record.get("t0")) is None:
                raise ValueError(f"{type(self).__name__} requires t0 records")
            pps = sorted(
                [pp for pp in photopoints if "LSST_DP" in pp.get("tag", [])],
                key=lambda pp: pp["body"]["midpointMjdTai"],
            )
            if not pps:
                return

            def _diasource_id(pp) -> int:
                body = pp.get("body", {})
                dsid = body.get("diaSourceId")
                if dsid is None:
                    raise KeyError(
                        f"No diaSourceId in photopoint body keys={list(body.keys())}"
                    )
                return int(dsid)

            if self.eligible == "last":
                candids = [_diasource_id(pps[-1])]
            elif self.eligible == "first":
                candids = [_diasource_id(pps[0])]
            elif self.eligible == "brightest":
                candids = [
                    _diasource_id(max(pps, key=lambda pp: pp["body"]["psfFlux"]))
                ]
            else:  # all
                candids = [_diasource_id(pp) for pp in pps]

            cutouts = {candid: self.get_cutout(candid) for candid in candids}

            if "extra" not in record or record["extra"] is None:
                record["extra"] = {self.__class__.__name__: cutouts}
            else:
                record["extra"][self.__class__.__name__] = cutouts



import os
import uuid
from threading import Event
from typing import Annotated, Any, Self

import confluent_kafka
from annotated_types import Gt, MinLen
from confluent_kafka.deserializing_consumer import DeserializingConsumer

from ampel.abstract.AbsContextManager import AbsContextManager
from ampel.base.AmpelUnit import AmpelUnit

from .KafkaAuthentication import KafkaAuthentication, SASLAuthentication


class KafkaConsumerBase(AbsContextManager, AmpelUnit):
    #: Address of Kafka broker
    bootstrap: str
    #: Optional authentication
    auth: None | KafkaAuthentication = None
    #: Topics to subscribe to
    topics: Annotated[list[str], MinLen(1)]
    #: Consumer group name
    group_name: None | str = None
    #: time to wait for messages before giving up, in seconds
    timeout: Annotated[int, Gt(0)] = 1
    #: environment variable to use as group.instance.id. If None, disable static membership
    instance_id_env_var: None | str = "HOSTNAME"
    #: extra configuration to pass to confluent_kafka.Consumer
    kafka_consumer_properties: dict[str, Any] = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        group_name = self.group_name if self.group_name else str(uuid.uuid1())
        if isinstance(self.auth, SASLAuthentication):
            group_name = f"{self.auth.username.get()}-{group_name}"

        config = (
            {
                "bootstrap.servers": self.bootstrap,
                "auto.offset.reset": "smallest",
                "enable.auto.commit": True,
                "enable.auto.offset.store": False,
                "auto.commit.interval.ms": 10000,
                "receive.message.max.bytes": 2**29,
                "enable.partition.eof": False,  # don't emit messages on EOF
                "error_cb": self._raise_errors,
                "group.id": group_name,
            }
            | (self.auth.librdkafka_config() if self.auth is not None else {})
            # allow process to restart without triggering a rebalance
            | (
                {"group.instance.id": os.getenv(self.instance_id_env_var)}
                if self.instance_id_env_var and os.getenv(self.instance_id_env_var)
                else {}
            )
            | self.kafka_consumer_properties
        )

        self._consumer = DeserializingConsumer(config)

        self._poll_interval = max((1, min((3, self.timeout))))
        self._poll_attempts = max((1, int(self.timeout / self._poll_interval)))

    def _raise_errors(self, exc: Exception) -> None:
        raise exc

    def _poll(self, stop: Event) -> confluent_kafka.Message | None:
        """
        Poll for a message, ignoring nonfatal errors
        """
        message = None
        # wake up occasionally to catch SIGINT
        for _ in range(self._poll_attempts):
            try:
                if stop.is_set() or (
                    message := self._consumer.poll(self._poll_interval)
                ):
                    break
            except confluent_kafka.KafkaError as exc:
                if exc.code() == confluent_kafka.KafkaError.UNKNOWN_TOPIC_OR_PART:
                    # ignore unknown topic messages
                    continue
                if exc.code() in (
                    confluent_kafka.KafkaError._TIMED_OUT,  # noqa: SLF001
                    confluent_kafka.KafkaError._MAX_POLL_EXCEEDED,  # noqa: SLF001
                ):
                    # bail on timeouts
                    return None
                raise
            except KeyboardInterrupt:
                stop.set()
        return message

    def __enter__(self) -> Self:
        self._consumer.subscribe(self.topics)
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._consumer.commit()
        self._consumer.unsubscribe()



import io

import fastavro
from confluent_kafka.serialization import Deserializer, SerializationContext

from .HttpSchemaRepository import parse_schema


class PlainAvroDeserializer(Deserializer):
    """
    Deserializer for static schemas
    """

    def __init__(self, avro_schema: dict | str):
        self._schema = parse_schema(avro_schema)

    def __call__(self, value: bytes, ctx: SerializationContext) -> dict:  # noqa: ARG002
        return fastavro.schemaless_reader(  # type: ignore[return-value]
            io.BytesIO(value),
            writer_schema=self._schema,
            reader_schema=None,
        )



from collections.abc import Generator, Iterable

from ampel.base.AmpelBaseModel import AmpelBaseModel


class StringFilter(AmpelBaseModel):
    include: None | set[str] = None
    exclude: None | set[str] = None

    def filter(self, fields: Iterable[str]) -> Generator[str, None, None]:
        for field in fields:
            if self.include is not None and field not in self.include:
                continue
            if self.exclude is not None and field in self.exclude:
                continue
            yield field


class FieldSelection(AmpelBaseModel):
    """
    Field selection for LSST data points
    """

    diaSource: None | StringFilter = None
    diaForcedSource: None | StringFilter = None
    diaObject: None | StringFilter = None
    diaNondetectionLimit: None | StringFilter = None

    @classmethod
    def default(cls) -> "FieldSelection":
        return FieldSelection(
            diaSource=StringFilter(
                include={
                    "band",
                    "dec",
                    "isNegative",
                    "midpointMjdTai",
                    "psfFlux",
                    "psfFluxErr",
                    "ra",
                    "snr",
                    "visit",
                    "diaSourceId",
                },
                exclude=None,
            ),
            diaForcedSource=StringFilter(
                include={
                    "band",
                    "dec",
                    "midpointMjdTai",
                    "psfFlux",
                    "psfFluxErr",
                    "ra",
                    "visit",
                    "diaForcedSourceId",
                },
                exclude=None,
            ),
            diaObject=StringFilter(
                include={
                    "dec",
                    "decErr",
                    "ra_dec_Cov",
                    "ra",
                    "radecMjdTai",
                    "raErr",
                    "diaObjectId",
                },
                exclude=None,
            ),
            diaNondetectionLimit=None,
        )



from pathlib import PurePosixPath
from urllib.parse import urlsplit, urlunsplit

import fastavro.schema
import requests
from fastavro.repository.base import (
    AbstractSchemaRepository,
    SchemaRepositoryError,
)
from fastavro.types import Schema

# Schema used until ~summer 2023
# DEFAULT_SCHEMA = "https://raw.githubusercontent.com/LSSTDESC/elasticc/c47fbd301b87f915c77ac0046d7845c68c306444/alert_schema/elasticc.v0_9.alert.avsc"
# Current default
DEFAULT_SCHEMA = "https://raw.githubusercontent.com/LSSTDESC/elasticc/main/alert_schema/elasticc.v0_9_1.alert.avsc"


class HttpSchemaRepostory(AbstractSchemaRepository):
    @classmethod
    def get_parts(cls, url: str) -> tuple[str, str, str]:
        parts = urlsplit(url)
        path = PurePosixPath(parts.path)
        return (
            urlunsplit(
                (
                    parts.scheme,
                    parts.netloc,
                    f"{path.parent}/",
                    parts.query,
                    parts.fragment,
                )
            ),
            path.stem,
            path.suffix,
        )

    @classmethod
    def load_schema(cls, url: str) -> Schema:
        base, schema, ext = cls.get_parts(url)
        return fastavro.schema.load_schema(schema, repo=cls(base, ext))

    def __init__(self, base_url: str, file_ext: str = ".avsc"):
        self.base_url = base_url
        self.file_ext = file_ext
        self.session = requests.Session()

    def load(self, name: str):
        try:
            response = self.session.get(f"{self.base_url}{name}{self.file_ext}")
            response.raise_for_status()
            return response.json()
        except Exception as error:
            raise SchemaRepositoryError(
                f"Failed to load '{name}' schema",
            ) from error


def parse_schema(schema_or_url: str | dict) -> Schema:
    if isinstance(schema_or_url, str):
        return HttpSchemaRepostory.load_schema(schema_or_url)
    return fastavro.schema.parse_schema(schema_or_url)



#!/usr/bin/env python
# File              : Ampel-LSST/ampel/lsst/alert/ElasticcAlertSupplier.py
# License           : BSD-3-Clause
# Author            : j nordin <jnordin@physik.hu-berlin.de>
# Date              : 9.6.2022
# Last Modified Date: 22.6.2022
# Last Modified By  : j nordin <jnordin@physik.hu-berlin.de>

import sys
from hashlib import blake2b
from typing import Literal, cast

from astropy.table import Table
from bson import encode

from ampel.alert.AmpelAlert import AmpelAlert
from ampel.alert.BaseAlertSupplier import BaseAlertSupplier
from ampel.protocol.AmpelAlertProtocol import AmpelAlertProtocol
from ampel.view.ReadOnlyDict import ReadOnlyDict


class ElasticcAlertSupplier(BaseAlertSupplier):
    """
    Iterable class that, for each lightcurve provided by the underlying alert_loader,
    returns an AmpelAlert instance.
    """

    # Override default
    deserialize: None | Literal["avro", "json"]

    def __next__(self) -> AmpelAlertProtocol:
        """
        :returns: a dict with a structure that AlertConsumer understands
        :raises StopIteration: when alert_loader dries out.
        :raises AttributeError: if alert_loader was not set properly before this method is called
        """

        lc = cast(Table, self._deserialize(next(self.alert_loader)))

        # Are these actually unique?
        sntype = lc.meta["sim_type_index"]  # Again do not know how consistent these are
        # Try to generate a unique stock. Not sure how well this will work.
        # Lets see whether we still get dubplicates...
        if lc.meta["ra"] == 0.0:  # Unclear why this happens
            stock = int(lc.meta["snid"] + "99")
        else:
            stock = int(lc.meta["snid"] + str(abs(lc.meta["ra"]) * 10000)[0:2])

        # Generate datapoints.
        # Would be much more efficient if dps generation was done in Loader?
        dps = []
        all_ids = b""
        df = lc.to_pandas()
        for _, row in df.iterrows():
            d = dict(row)
            # Generate point id through hash
            d_hash = blake2b(encode(d), digest_size=7).digest()
            # We will call this a diaSourceId if it is associated with a detection
            # Only these are asigned and obsvered magn
            if d["cause_alert"]:
                d["diaSourceId"] = int.from_bytes(d_hash, byteorder=sys.byteorder)
            else:
                # Take this to be forced photometry
                d["diaForcedSourceId"] = int.from_bytes(d_hash, byteorder=sys.byteorder)
            all_ids += d_hash
            dps.append(ReadOnlyDict(d))

        # Create alert id through hash of obs dates
        alert_id = int.from_bytes(  # alert id
            blake2b(all_ids, digest_size=7).digest(), byteorder=sys.byteorder
        )

        # Add meta as content of diaObject + alert info
        d = dict(lc.meta)
        d["alertId"] = alert_id
        # Is it an issue that diaObjectId is the same as stock?
        d["diaObjectId"] = stock
        dps.append(ReadOnlyDict(d))

        return AmpelAlert(
            id=alert_id,  # alert id
            stock=stock,  # internal ampel id
            datapoints=tuple(dps),
            tag=["GENTYPE" + str(sntype)],
        )



import io

import fastavro
from confluent_kafka.serialization import SerializationContext, Serializer

from .HttpSchemaRepository import parse_schema


class PlainAvroSerializer(Serializer):
    """
    Serializer for static schemas
    """

    def __init__(self, avro_schema: dict | str):
        self._schema = parse_schema(avro_schema)

    def __call__(
        self,
        value: dict | None,
        ctx: SerializationContext,  # noqa: ARG002
    ) -> bytes | None:
        if value is None:
            return None
        with io.BytesIO() as buf:
            fastavro.schemaless_writer(buf, self._schema, value)
            return buf.getvalue()



#!/usr/bin/env python
# File:                Ampel-LSST/ampel/lsst/t2/T2GetAlertJournal.py
# License:             BSD-3-Clause
# Author:              Marcus Fennner <mf@physik.hu-berlinn.de>
# Date:                31.03.2022
# Last Modified Date:  31.03.2022
# Last Modified By:    Marcus Fennner <mf@physik.hu-berlinn.de>


from ampel.abstract.AbsStockT2Unit import AbsStockT2Unit
from ampel.content.StockDocument import StockDocument
from ampel.types import UBson


class T2GetAlertJournal(AbsStockT2Unit):
    """
    Get the alert journal of an LSST stock
    """

    def process(self, stock_doc: StockDocument) -> UBson:
        return [journal for journal in stock_doc["journal"] if journal["tier"] == 0]



#!/usr/bin/env python
# File              : Ampel-LSST/ampel/lsst/alert/LSSTAlertSupplier.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 20.04.2021
# Last Modified Date: 21.03.2022
# Last Modified By  : Marcus Fenner <mf@physik.hu-berlin.de>

from collections.abc import Generator, Iterator
from itertools import chain
from typing import Literal

from ampel.alert.AmpelAlert import AmpelAlert
from ampel.alert.BaseAlertSupplier import BaseAlertSupplier
from ampel.protocol.AmpelAlertProtocol import AmpelAlertProtocol
from ampel.view.ReadOnlyDict import ReadOnlyDict


class DIAObjectMissingError(Exception):
    """
    Raised when there is no DIAObject in the alert
    """

    ...


# Translate 2022-era field names to schema 7.x
_field_upgrades: dict[str, str] = {
    "midPointTai": "midpointMjdTai",
    "psFlux": "psfFlux",
    "psFluxErr": "psfFluxErr",
    "filterName": "band",
    "decl": "dec",
    "ccdVisitId": "visit",
}


class LSSTAlertSupplier(BaseAlertSupplier):
    """
    Iterable class that, for each alert payload provided by the underlying alert_loader,
    returns an AmpelAlert instance.
    """

    # Override default
    deserialize: None | Literal["avro", "json"] = "avro"

    max_history: float = float("inf")

    alert_identifier: Literal["diaSourceId", "alertId"] = "diaSourceId"

    @staticmethod
    def _shape_dp(d: dict) -> ReadOnlyDict:
        return ReadOnlyDict({_field_upgrades.get(k, k): v for k, v in d.items()})

    @classmethod
    def _get_sources(
        cls, alert: dict, max_history: float
    ) -> Generator[dict, None, None]:
        """
        yield one photometric point per visit, preferring forced photometry to
        difference imaging and taking upper limits as a last resort
        """
        diaSource = cls._shape_dp(alert["diaSource"])
        yield diaSource

        visits: set[int] = set()
        t0 = diaSource["midpointMjdTai"] - max_history
        for dp in map(
            cls._shape_dp,
            chain(
                alert.get("prvDiaForcedSources") or (),
                alert.get("prvDiaSources") or (),
                alert.get("diaNondetectionLimit") or (),
            ),
        ):
            if (visit := dp["visit"]) not in visits and dp["midpointMjdTai"] >= t0:
                yield dp
                visits.add(visit)

    @classmethod
    def _shape(
        cls,
        d: dict,
        max_history: float = float("inf"),
        alert_identifier: Literal["diaSourceId", "alertId"] = "diaSourceId",
    ) -> AmpelAlertProtocol:
        if diaObject := d.get("diaObject"):
            dps = (
                *cls._get_sources(d, max_history=max_history),
                cls._shape_dp(diaObject),
            )
            # Add base alert information to extras field
            extras = {}
            for alertprop in ["observation_reason", "target_name"]:
                if val := d.get(alertprop):
                    extras[alertprop] = val
            if kafka := d.get("__kafka"):
                extras["kafka"] = kafka
            return AmpelAlert(
                id=d[
                    alert_identifier
                ],  # ID of the triggering DiaSource - use as alert id?
                stock=diaObject["diaObjectId"],  # internal ampel id
                datapoints=dps,
                extra=extras if len(extras) > 0 else None,
            )
        raise DIAObjectMissingError

    def acknowledge(self, alerts: Iterator[AmpelAlertProtocol]) -> None:
        # invert transformation applied in _shape()
        self.alert_loader.acknowledge(
            {"__kafka": alert.extra["kafka"]}  # type: ignore[misc]
            for alert in alerts
            if alert.extra and "kafka" in alert.extra
        )

    def __next__(self) -> AmpelAlertProtocol:
        """
        :returns: a dict with a structure that AlertConsumer understands
        :raises StopIteration: when alert_loader dries out.
        :raises AttributeError: if alert_loader was not set properly before this method is called
        """
        while True:
            d = self._deserialize(next(self.alert_loader))

            try:
                return self._shape(d, self.max_history, self.alert_identifier)
            except DIAObjectMissingError:
                # silently skip over SSObjects
                continue



from functools import cached_property
from typing import Any

from confluent_kafka.schema_registry import SchemaRegistryClient
from confluent_kafka.schema_registry.avro import (
    AvroDeserializer,
    AvroSerializer,
)
from confluent_kafka.serialization import Deserializer, Serializer

from ampel.base.AmpelBaseModel import AmpelBaseModel

from .HttpSchemaRepository import DEFAULT_SCHEMA
from .PlainAvroDeserializer import PlainAvroDeserializer
from .PlainAvroSerializer import PlainAvroSerializer


class SchemaRegistryURL(AmpelBaseModel):
    registry: str
    subject: str | None = None

    @cached_property
    def _schema_registry_client(self) -> SchemaRegistryClient:
        return SchemaRegistryClient({"url": self.registry})

    def serializer(self) -> Serializer:
        # Require that schemas are pre-registered
        conf: dict[str, Any] = {
            "auto.register.schemas": False,
            "use.latest.version": True,
        }
        if self.subject is not None:
            conf["subject.name.strategy"] = lambda *args: self.subject
        return AvroSerializer(self._schema_registry_client, conf=conf)

    def deserializer(self) -> Deserializer:
        return AvroDeserializer(self._schema_registry_client)


class StaticSchemaURL(AmpelBaseModel):
    root_url: str = DEFAULT_SCHEMA

    def serializer(self) -> Serializer:
        return PlainAvroSerializer(self.root_url)

    def deserializer(self) -> Deserializer:
        return PlainAvroDeserializer(self.root_url)


AvroSchema = SchemaRegistryURL | StaticSchemaURL



#!/usr/bin/env python
# File:                Ampel-LSST/ampel/lsst/t0/SimpleLSSTFilter.py
# License:             BSD-3-Clause
# Author:              m. giomi <matteo.giomi@desy.de>
# Date:                06.06.2018
# Last Modified Date:  24.03.2022
# Last Modified By:    Marcus Fenner <mf@physik.hu-berlin.de>

from typing import Any

import numpy as np
from astropy.coordinates import SkyCoord
from astropy.table import Table

from ampel.abstract.AbsAlertFilter import AbsAlertFilter
from ampel.protocol.AmpelAlertProtocol import AmpelAlertProtocol
from ampel.ztf.base.CatalogMatchUnit import CatalogMatchUnit


class SimpleLSSTFilter(CatalogMatchUnit, AbsAlertFilter):
    """
    General-purpose filter. It selects alerts based on:
    * numper of previous detections
    * distance to known SS objects
    * detection of proper-motion and paralax for coincidence sources in GAIA DR2

    """

    # History
    min_ndet: int  # number of previous detections
    min_tspan: float  # minimum duration of alert detection history [days]
    max_tspan: float  # maximum duration of alert detection history [days]
    min_archive_tspan: float = 0.0  # minimum duration of alert detection history [days]
    max_archive_tspan: float = (
        10**5.0
    )  # maximum duration of alert detection history [days]

    # Astro
    #    min_sso_dist: float  # distance to nearest solar system object [arcsec]
    min_gal_lat: (
        float  # minium distance from galactic plane. Set to negative to disable cut.
    )

    # Gaia
    gaia_rs: float  # search radius for GAIA DR2 matching [arcsec]
    gaia_pm_signif: (
        float  # significance of proper motion detection of GAIA counterpart [sigma]
    )
    gaia_plx_signif: (
        float  # significance of parallax detection of GAIA counterpart [sigma]
    )
    gaia_veto_gmag_min: (
        float  # min gmag for normalized distance cut of GAIA counterparts [mag]
    )
    gaia_veto_gmag_max: (
        float  # max gmag for normalized distance cut of GAIA counterparts [mag]
    )
    gaia_excessnoise_sig_max: float  # maximum allowed noise (expressed as significance) for Gaia match to be trusted.

    def post_init(self):
        # feedback
        for k in self.__annotations__:
            self.logger.info(f"Using {k}={getattr(self, k)}")

        # To make this tenable we should create this list dynamically depending on what entries are required
        # by the filter. Now deciding not to include drb in this list, eg.
        self.keys_to_check = (
            "midpointMjdTai",
            "ra",
            "decl",
        )

    def _alert_has_keys(self, photop) -> bool:
        """
        check that given photopoint contains all the keys needed to filter
        """
        for el in self.keys_to_check:
            if el not in photop:
                self.logger.info(None, extra={"missing": el})
                return False
            if photop[el] is None:
                self.logger.info(None, extra={"isNone": el})
                return False
        return True

    def get_galactic_latitude(self, transient):
        """
        compute galactic latitude of the transient
        """
        coordinates = SkyCoord(transient["ra"], transient["decl"], unit="deg")
        return coordinates.galactic.b.deg

    def is_star_in_gaia(self, transient: dict[str, Any]) -> bool:
        """
        match tranient position with GAIA DR2 and uses parallax
        and proper motion to evaluate star-likeliness
        returns: True (is a star) or False otehrwise.
        """

        srcs = self.cone_search_all(
            transient["ra"],
            transient["decl"],
            [
                {
                    "name": "GAIADR2",
                    "use": "catsHTM",
                    "rs_arcsec": self.gaia_rs,
                    "keys_to_append": [
                        "Mag_G",
                        "PMRA",
                        "ErrPMRA",
                        "PMDec",
                        "ErrPMDec",
                        "Plx",
                        "ErrPlx",
                        "ExcessNoiseSig",
                    ],
                }
            ],
        )[0]

        if srcs:
            gaia_tab = Table(
                [
                    {k: np.nan if v is None else v for k, v in src["body"].items()}
                    for src in srcs
                ]
            )

            # compute distance
            gaia_tab["DISTANCE"] = [src["dist_arcsec"] for src in srcs]
            gaia_tab["DISTANCE_NORM"] = (
                1.8 + 0.6 * np.exp((20 - gaia_tab["Mag_G"]) / 2.05)
                > gaia_tab["DISTANCE"]
            )
            gaia_tab["FLAG_PROX"] = [
                x["DISTANCE_NORM"]
                and self.gaia_veto_gmag_min <= x["Mag_G"] <= self.gaia_veto_gmag_max
                for x in gaia_tab
            ]

            # check for proper motion and parallax conditioned to distance
            gaia_tab["FLAG_PMRA"] = (
                abs(gaia_tab["PMRA"] / gaia_tab["ErrPMRA"]) > self.gaia_pm_signif
            )
            gaia_tab["FLAG_PMDec"] = (
                abs(gaia_tab["PMDec"] / gaia_tab["ErrPMDec"]) > self.gaia_pm_signif
            )
            gaia_tab["FLAG_Plx"] = (
                abs(gaia_tab["Plx"] / gaia_tab["ErrPlx"]) > self.gaia_plx_signif
            )

            # take into account precison of the astrometric solution via the ExcessNoise key
            gaia_tab["FLAG_Clean"] = (
                gaia_tab["ExcessNoiseSig"] < self.gaia_excessnoise_sig_max
            )

            # select just the sources that are close enough and that are not noisy
            gaia_tab = gaia_tab[gaia_tab["FLAG_PROX"]]
            gaia_tab = gaia_tab[gaia_tab["FLAG_Clean"]]

            # among the remaining sources there is anything with
            # significant proper motion or parallax measurement
            if (
                any(gaia_tab["FLAG_PMRA"] == True)  # noqa: E712
                or any(gaia_tab["FLAG_PMDec"] == True)  # noqa: E712
                or any(gaia_tab["FLAG_Plx"] == True)  # noqa: E712
            ):
                return True

        return False

    # Override
    def process(self, alert: AmpelAlertProtocol) -> None | bool:
        """
        Mandatory implementation.
        To exclude the alert, return *None*
        To accept it, either return
        * self.on_match_t2_units
        * or a custom combination of T2 unit names
        """

        # CUT ON THE HISTORY OF THE ALERT
        #################################

        pps = [el for el in alert.datapoints if el.get("diaSourceId") is not None]
        if len(pps) < self.min_ndet:
            # self.logger.debug("rejected: %d photopoints in alert (minimum required %d)"% (npp, self.min_ndet))
            self.logger.info(None, extra={"nDet": len(pps)})
            return None

        # cut on length of detection history
        detections_jds = [el["midpointMjdTai"] for el in pps]
        det_tspan = max(detections_jds) - min(detections_jds)
        if not (self.min_tspan <= det_tspan <= self.max_tspan):
            # self.logger.debug("rejected: detection history is %.3f d long, \
            # requested between %.3f and %.3f d"% (det_tspan, self.min_tspan, self.max_tspan))
            self.logger.info(None, extra={"tSpan": det_tspan})
            return None

        # IMAGE QUALITY CUTS
        ####################

        latest = alert.datapoints[0]
        # latest = sorted(alert.datapoints, key=lambda x:x["jd"], reverse=True)[0]
        if not self._alert_has_keys(latest):
            return None

        # cut on galactic latitude
        b = self.get_galactic_latitude(latest)
        if abs(b) < self.min_gal_lat:
            # self.logger.debug("rejected: b=%.4f, too close to Galactic plane (max allowed: %f)."% (b, self.min_gal_lat))
            self.logger.info(None, extra={"galPlane": abs(b)})
            return None

        # check with gaia
        if self.gaia_rs > 0 and self.is_star_in_gaia(latest):
            self.logger.debug(
                f"rejected: within {self.gaia_rs:.2f} arcsec from a GAIA start (PM of PLX)"
            )
            self.logger.info(None, extra={"gaiaIsStar": True})
            return None

        # self.logger.debug("Alert %s accepted. Latest pp ID: %d"%(alert.tran_id, latest['candid']))
        self.logger.debug("Alert accepted", extra={"latestPpId": latest["diaSourceId"]})

        # for key in self.keys_to_check:
        # 	self.logger.debug("{}: {}".format(key, latest[key]))

        return True



from typing import Any, Literal

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.secret.NamedSecret import NamedSecret


class SASLAuthentication(AmpelBaseModel):
    protocol: Literal["SASL_PLAINTEXT", "SASL_SSL"] = "SASL_PLAINTEXT"
    mechanism: Literal["PLAIN", "SCRAM-SHA-256", "SCRAM-SHA-512"] = "SCRAM-SHA-512"
    username: NamedSecret[str]
    password: NamedSecret[str]

    def librdkafka_config(self) -> dict[str, Any]:
        return {
            "security.protocol": self.protocol,
            "sasl.mechanism": self.mechanism,
            "sasl.username": self.username.get(),
            "sasl.password": self.password.get(),
        }
